@@ -1084,10 +1084,14 @@ static dispatch_semaphore_t _datawaiting;
 #else
 static sem_t *datawaiting = NULL;
 static sem_t _datawaiting;
-static int sema_wait(sem_t *sem, int timeout) {
+static int sema_wait(sem_t *sem, long int timeout) {
 	struct timespec wait;
-	wait.tv_sec = 0;
-	wait.tv_nsec = timeout;
+	clock_gettime(CLOCK_REALTIME, &wait);
+	wait.tv_nsec += timeout;
+	if (wait.tv_nsec >= 1000000000) {
+		wait.tv_sec++;
+		wait.tv_nsec -= 1000000000;
+	}
 	return sem_timedwait(sem, &wait);
 }
 #endif
@@ -1130,7 +1134,7 @@ dispatch_runner(void *arg)
 				if (__sync_bool_compare_and_swap(&(conn->datawaiting), 0, 0))
 				{
 					ufds[fds].fd = conn->sock;
-					ufds[fds].events = POLLIN | POLLERR;
+					ufds[fds].events = POLLIN;
 					fds++;
 				}
 			}
@@ -1153,7 +1157,7 @@ dispatch_runner(void *arg)
 			for (f = 0; f < fds; f++) {
 				if (f < cfds) {
 					/* connection has data available */
-					if (ufds[f].revents & (POLLIN | POLLERR)) {
+					if (ufds[f].revents & (POLLIN | POLLERR | POLLHUP)) {
 						pthread_rwlock_rdlock(&connectionslock);
 						for (c = 0; c < connectionslen; c++) {
 							conn = &(connections[c]);
@@ -1179,14 +1183,12 @@ dispatch_runner(void *arg)
 						struct sockaddr addr;
 						socklen_t addrlen = sizeof(addr);
 
-						if ((client = accept(ufds[f].fd, &addr, &addrlen)) < 0)
-						{
-							logerr("dispatch: failed to "
-									"accept() new connection: %s\n",
-									strerror(errno));
-							dispatch_check_rlimit_and_warn();
-							continue;
-						}
+						/* check if this connection belongs to an
+						 * existing listener, this is in particular of
+						 * importance when a listener was removed,
+						 * because thay may have interleaved here, and
+						 * we might get a POLLIN event for a closed
+						 * socket */
 						pthread_rwlock_rdlock(&listenerslock);
 						for (c = 0; c < MAX_LISTENERS; c++) {
 							if (listeners[c] == NULL)
@@ -1201,9 +1203,22 @@ dispatch_runner(void *arg)
 						}
 						pthread_rwlock_unlock(&listenerslock);
 						if (c == MAX_LISTENERS) {
-							logerr("dispatch: could not find listener for "
-									"socket, rejecting connection\n");
-							close(client);
+							/* be silent about this, because this
+							 * happens in tests, and isn't ever possible
+							 * in normal life */
+							tracef("dispatch: could not find listener for "
+									"socket, rejecting connection, fd=%d\n",
+									ufds[f].fd);
+							continue;
+						}
+
+						if ((client = accept(ufds[f].fd, &addr, &addrlen)) < 0)
+						{
+							logerr("dispatch: failed to "
+									"accept() new connection on %s:%d: %s\n",
+									listeners[c]->ip, listeners[c]->port,
+									strerror(errno));
+							dispatch_check_rlimit_and_warn();
 							continue;
 						}
 						if (dispatch_addconnection(client, listeners[c]) == -1)
@@ -1265,7 +1280,7 @@ dispatch_runner(void *arg)
 				/* wait a bit, but immediately spurt into action if
 				 * there's data available */
 				if (sema_wait(datawaiting,  /* 700ms - 999ms */
-							(700 + (rand() % 1000)) * 1000000) == 0)
+							(700 + (rand() % 300)) * 1000000) == 0)
 					tracef("dispatcher %d woken up\n", self->id);
 				gettimeofday(&stop, NULL);
 				__sync_add_and_fetch(&(self->sleeps), timediff(start, stop));
